/*
 * Device Driver for DataTranslation DT2801
 *
 * Analog-Digital-Converter subsystem
 *
 */

#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/malloc.h>
#include <linux/timer.h>

#include <asm/io.h>
#include <asm/dma.h>
#include <asm/segment.h>
#include <asm/system.h>

#include <linux/wrapper.h> /* for mem_map_reserve() ... */

#ifndef MODULE
#error "adc.c must be compiled as a module"
#endif

#include "dt2801.h"
#include "adc.h"


/* routines for faked driver operation, allows to test the driver
   when no hardware is present ( fake=1 as argument, to activate) */


static void adc_start_timer(ADCP* adcp);

static void adc_timer_interrupt(unsigned long bp)
{
	ADCP* adcp = (ADCP*) bp;
	Dt2801_t* boardPtr = adcp->board;

        del_timer(&boardPtr->tl);
        if (boardPtr->fake != 0) {
                short *b = (short *) adcp->dmabuf;

                b[boardPtr->fake_counter] = boardPtr->fake_counter%5;
                boardPtr->fake_counter++;
		adc_start_timer(adcp);
        }
}

static void adc_start_timer(ADCP* adcp)
{
	Dt2801_t* boardPtr = adcp->board;

        if (boardPtr->fake_counter >= adcp->conv) {
		boardPtr->running = 0;
                return;
        }
	if (boardPtr->stop_req != 0) {
		boardPtr->running = 0;
                return;
	}

        init_timer(&boardPtr->tl);
        boardPtr->tl.expires = jiffies + 20;
	boardPtr->tl.data = (unsigned long) adcp;
        boardPtr->tl.function = adc_timer_interrupt;
        add_timer(&boardPtr->tl);
}

/* end of fake routines */

static int allocate_mmapadc_buffer(ADCP* adcp, int bufsize)
{
	int sz, size, i;
	char* start_addr, *end_addr;

        if (bufsize <= 0) {
                printk("dt2801_adc: allocating DMA-buffer,"
                       " illegal bufsize %d\n", bufsize);
                return -(EINVAL);
        }
        if (bufsize > DT2801_MAX_DMA_SIZE) {
                printk("dt2801_adc: Requested DMA buffer size,"
                       " %d, too large\n", bufsize );
                return -(EINVAL);
        }

        /* reuse the buffer if it is large enough */
        if (bufsize <= adcp->dmabufsize ) {
                return 0;
        }
 
        if (adcp->dmabuf != NULL) {
                printk("dt2801_adc: freeing dmabuf\n");
                kfree(adcp->dmabuf);
                adcp->dmabufsize = 0;
                adcp->dmabuf = NULL;
        }

        /* Allocate buffer for DMA transfers */

	/* mmapable */
	/* Taken from the sounddriver. It's not clear to
	   me what the requirements for these pages are
	   to make the mmap work, but this one does. */

	for (sz = 0, size = PAGE_SIZE; size < bufsize; sz++, size <<= 1);

	printk("dt2801: sz = %d\n", sz);

        adcp->dmabufsize = PAGE_SIZE * (1 << sz);
	start_addr =  __get_free_pages(GFP_KERNEL, sz, MAX_DMA_ADDRESS);
	if (start_addr == NULL) {
		printk("dt2801: __get_free_pages failed\n");
		return  -(ENOMEM);
	}
	end_addr = start_addr + adcp->dmabufsize - 1;
	for (i = MAP_NR(start_addr); i <= MAP_NR(end_addr); i++) {
		mem_map_reserve (i);
	}
	adcp->dmabuf = start_addr;
	adcp->sz     = sz;

        return 0;
}

static int allocate_adc_buffer(ADCP* adcp, int bufsize)
{
        if (bufsize <= 0) {
                printk("dt2801_adc: allocating DMA-buffer,"
                       " illegal bufsize %d\n", bufsize);
                return -(EINVAL);
        }
        if (bufsize > DT2801_MAX_DMA_SIZE) {
                printk("dt2801_adc: Requested DMA buffer size,"
                       " %d, too large\n", bufsize );
                return -(EINVAL);
        }

        /* reuse the buffer if it is large enough */
        if (bufsize <= adcp->dmabufsize ) {
                return 0;
        }
 
        if (adcp->dmabuf != NULL) {
                printk("dt2801_adc: freeing dmabuf\n");
                kfree(adcp->dmabuf);
                adcp->dmabufsize = 0;
                adcp->dmabuf = NULL;
        }

        /* Allocate buffer for DMA transfers */
        adcp->dmabuf = kmalloc( bufsize, GFP_KERNEL | GFP_DMA );
        if (adcp->dmabuf == NULL) {
                printk( "dt2801_adc: Unable to allocate DMA buffer\n" );
                return -(ENOMEM);
        }
        adcp->dmabufsize = bufsize;

        return 0;
}

static void clear_adc_buffer(ADCP* adcp)
{
        int i;

        for (i=0; i<adcp->dmabufsize; i++) {
                adcp->dmabuf[i] = 0xf000;
        }
}

/* Find the first 2 byte value in the buffer, that is not yet
   complete. Only works with 12 bit ADC, not with 16 bit*/
static int get_currnt(ADCP* adcp)
{
        int i;

	/* This tests whether the highbyte still has its high nibble
	   set to 0xf, as it was done by 'clear_adc_buffer' This
	   works only on a little endian machine. This is necessary
	   to see, if the 2 bytes (12 bit significant) of the
	   conversion have both arrived. */
	for (i=adcp->cnt; i<adcp->conv; i++) {
		if (adcp->dmabuf[i] & 0xf000) {
                        break;
                }
        }

        if (i==0) {
		adcp->cnt = i;
	} else if (i == adcp->conv) {
		adcp->cnt = i;
	} else {
		adcp->cnt = i-1;
	}
        return adcp->cnt;
}

static int dt2801_set_adc_params(ADCP* adcp, Dt2801_ADCP_t *adcpar)
{                
        int stat;
	Dt2801_t* boardPtr = adcp->board;

        dt2801_writecommand(boardPtr, DT_C_SET_AD);

        dt2801_writedata(boardPtr, adcpar->gain);
        dt2801_writedata(boardPtr, adcpar->startch);
        dt2801_writedata(boardPtr, adcpar->endch);
        dt2801_writedata2(boardPtr, adcpar->conv);

        stat = dt2801_checkforerror(boardPtr);
        if (stat!=0) {
                return stat;
        }

        adcp->last = 0;
        adcp->gain = adcpar->gain;
        adcp->startch = adcpar->startch;
        adcp->endch = adcpar->endch;
        adcp->conv = adcpar->conv;
        adcp->cnt = 0;
        
	/*
	stat = allocate_mmapadc_buffer(adcp, adcp->conv*sizeof(short));
	if (stat != 0) {
		return stat;
	}
	*/
	stat = allocate_adc_buffer(adcp, adcp->conv*sizeof(short));
	if (stat != 0) {
		return stat;
	}
	
        return 0;
}

static int dt2801_adc_ioctl_a(struct inode *inode, struct file *filp,
			      unsigned int cmd, unsigned long arg)
{
	ADCP* adcp = (ADCP*) filp->private_data;
	Dt2801_t* boardPtr = adcp->board;
	ADC_state* adc_cf = &boardPtr->adc_state;

        switch (cmd) {
        case DT2801_IOCTL_ADC_GET_POLARITY: {
                int stat;
                
                stat = verify_area(VERIFY_WRITE, (char *)arg, sizeof(int));
                if (stat != 0) {
                        return stat;
                }
                put_user(adc_cf->polarity, (int*)arg);
        }
        break;
	case DT2801_IOCTL_ADC_SET_GAIN: {
		int gain = (int) arg;
		adcp->gain = gain;
	}
	break;
	case DT2801_IOCTL_ADC_SETEXTTRIGGER: {
		int exttrig = (int) arg;
		adcp->exttrigger = exttrig;
	}
	break;
        default: {
                return dt2801_do_ioctl(inode, filp, cmd, arg, boardPtr);
        }
        }

        return 0;
}

static int dt2801_adc_ioctl(struct inode *inode, struct file *filp,
                            unsigned int cmd, unsigned long arg)
{
	ADCP* adcp = (ADCP*) filp->private_data;
	Dt2801_t* boardPtr = adcp->board;
	int stat;

        switch (cmd) {
        case DT2801_IOCTL_ADC_SET_PARAMS: {
                Dt2801_ADCP_t uadcp;

                memcpy_fromfs(&uadcp, (char*)arg, sizeof(uadcp));
                adcp->startch = uadcp.startch;
                adcp->endch = uadcp.endch;
                adcp->gain = uadcp.gain;
                adcp->conv = uadcp.conv;
                stat = dt2801_set_adc_params(adcp, &uadcp);
		clear_adc_buffer(adcp);
		return stat;
        }
        break;
        
        /* ioctl SET_ADC_PARAMS has to be called before every call to
           BLOCKSTART */
        case DT2801_IOCTL_ADC_BLOCKSTART: {
                if (boardPtr->fake != 0) {
                        /* Start timer to fill the buffer */
                        if (boardPtr->running == 1) {
                                printk("dt2801: can't start, still running\n");
                                return -1;
                        }
                        clear_adc_buffer(adcp);
                        adcp->cnt = 0;
                
                        boardPtr->fake_counter = 0;
			boardPtr->running = 1;
			boardPtr->stop_req = 0;
                        adc_start_timer(adcp);
                } else {
                        clear_adc_buffer(adcp);
                        adcp->cnt = 0;

                        /* no 'printk's within cli()/sti() sections */
                        cli();
                        disable_dma(boardPtr->dmachannel);
                        clear_dma_ff( boardPtr->dmachannel );
                        set_dma_addr( boardPtr->dmachannel, 
                                      (unsigned long)adcp->dmabuf );
                        set_dma_count( boardPtr->dmachannel,
                                       adcp->dmabufsize );
                        set_dma_mode( boardPtr->dmachannel, DMA_MODE_READ );
                        enable_dma(boardPtr->dmachannel);
                        sti();
                        
                        if (adcp->exttrigger == 1) {
                                dt2801_writecommand(boardPtr,
                                                    DT_C_READ_AD | DT_MOD_DMA
                                                    | DT_MOD_EXTTRIG);
                        } else {
                                dt2801_writecommand(boardPtr,
                                                    DT_C_READ_AD | DT_MOD_DMA);
                        }
                        if (dt2801_checkforerror(boardPtr) != 0) {
                                printk("dt2801: blockstart-dma not ok\n");
                                return -1;
                        }
                        adcp->running = 1;
                }
                
        }
        break;
        case DT2801_IOCTL_ADC_BLOCKSTOP: {
		dt2801_do_stop(boardPtr);
		/* Stopping a command causes the error-flag to be set,
		   will have to clear it !! */
		dt2801_clearerror(boardPtr);
		if (boardPtr->fake == 1) {
			boardPtr->stop_req = 1;
		} else {
			disable_dma(boardPtr->dmachannel);
		}
	}
        break;
        case DT2801_IOCTL_ADC_RESIDUE: {
                int stat, residue;
                
                stat = verify_area(VERIFY_WRITE, (char *)arg, sizeof(int));
                
                if ( stat != 0) {
                        return stat;
                }
                
		cli();
                residue = get_dma_residue(boardPtr->dmachannel);
		sti();
                
                put_user(residue, (int*)arg);
                
        }
        break;
	case DT2801_IOCTL_ADC_SETEXTCLOCK: {
		int extcl = (int) arg;
		adcp->extclock = extcl;
	}
	break;
	case DT2801_IOCTL_ADC_SETCONTINOUS: {
		int cont_flag;
		cont_flag = (int) arg;
	
		adcp->continous = cont_flag;
	}
	break;
        default: {
                return dt2801_adc_ioctl_a(inode, filp, cmd, arg);
        }
	}

    
        return 0;
}

static int adc_read_imm(struct inode *inode, struct file *filp,
                           char* buffer, int count)
{
	ADCP* adcp = (ADCP*) filp->private_data;
	Dt2801_t* boardPtr = adcp->board;

        int data;
        int stat;


        if (count != 4) {
		return -EINVAL;
        }

        stat = dt2801_writecommand(boardPtr, DT_C_READ_ADIM);
        dt2801_writedata(boardPtr, adcp->gain);
        dt2801_writedata(boardPtr, adcp->startch);
        stat = dt2801_readdata2(boardPtr, &data);

        if (stat != 0) {
             printk("dt2801: stat = %d\n", stat);
             return -EIO;
        }

        count = sizeof(data);
        if ((stat = verify_area(VERIFY_WRITE, buffer, count)) != 0) {
                return stat;
        }
        memcpy_tofs(buffer, &data, count);

        return count;
}

static int adc_use_residue = 0;

static int adc_read_block(struct inode *inode, struct file *filp,
                          char* buffer, int count)
{
	ADCP* adcp = (ADCP*) filp->private_data;
	Dt2801_t* boardPtr = adcp->board;

        int residue;
        int d, stat;


        if (adcp->last == adcp->conv) {
                return -(EIO); /* eof ? */
        }
                
        if (boardPtr->fake != 0) {
                adcp->currnt = get_currnt(adcp);
        } else {
		if (adc_use_residue == 1) {
			cli();
			residue = get_dma_residue(boardPtr->dmachannel);
			/* only transfer complete 2-byte values  */
			adcp->currnt = adcp->conv - ((residue+1)/sizeof(short));
			sti();
		} else {
			adcp->currnt = get_currnt(adcp);
		}
        }
        
        /* no data has arrived yet, generate error */
        if (adcp->currnt  == 0) {
                return -1;
        }
        
        
        d = adcp->currnt - adcp->last;
        if (d > count) {
                d = count;
        }

	/*printk("c l d %d  %d  %d\n", adcp->currnt, adcp->last, d);*/
         
        if (d < 0) {
		printk("d <0\n");
                return -(EINVAL);
        }
        
        if ((stat = verify_area(VERIFY_WRITE, buffer, d*sizeof(short))) != 0) {
                return stat;
        }
        
        memcpy_tofs(buffer, adcp->dmabuf+(adcp->last), d*sizeof(short));
        
        adcp->last+=d;
        
        return d*sizeof(short);
}

static int dt2801_adc_mmap(struct inode *inode, struct file *filp,
			   struct vm_area_struct *vma)
{
	ADCP* adcp = (ADCP*) filp->private_data;

	printk("dt2801: mmap\n");
        if (adcp->dmabuf == NULL) { /* nothing to map */
		return -1;
        }

	printk("dt2801: offset = %lx\n", vma->vm_offset);

	if (remap_page_range(vma_get_start(vma),
			     virt_to_phys(adcp->dmabuf),/*XXX fake-mode?*/
			     vma_get_end(vma) - vma_get_start(vma),
			     vma_get_page_prot(vma))) {
		printk("dt2801: adc: remap bad\n");
		return -EAGAIN;
	}

	vma_set_inode(vma, inode);
	inode_inc_count(inode);

	printk("dt2801: adc remap ok\n");

	return 0;
}

int dt2801_adc_reset(Dt2801_t* boardPtr)
{
	ADC_state* adc_cf = &boardPtr->adc_state;

        adc_cf->gain  = -1;
        adc_cf->startch  = -1;
        adc_cf->endch  = -1;
        adc_cf->conv  = -1;

        return 0;
}

/* May be called even if 'init' has not  !!!!!! */
int dt2801_adc_cleanup(Dt2801_t* boardPtr)
{

        return 0;
}

/* May be given as parameters at load-time */
static char* adc_polarity[] = {
	"unipolar", "unipolar", "unipolar", "unipolar"
};
static char* adc_inputmode[] = {
	"differential", "differential", "differential", "differential"
};

void dt2801_adc_info(Dt2801_t* boardPtr)
{
	ADC_state* adc_cf = &boardPtr->adc_state;
        int bid = boardPtr->boardid;

	printk("dt2801:(%d) ADC: %s, %s\n", bid,
	       adc_polarity[bid], adc_inputmode[bid]);

}

int dt2801_adc_init(Dt2801_t* boardPtr)
{
	ADC_state* adc_cf = &boardPtr->adc_state;
        int bid = boardPtr->boardid;

        adc_cf->bits  = 12;

        if (!strcmp(adc_polarity[bid], "unipolar")) {
		adc_cf->polarity  = DT2801_ADC_UNIPOLAR;
	} else if (!strcmp(adc_polarity[bid], "bipolar")) {
		adc_cf->polarity  = DT2801_ADC_BIPOLAR;
	} else {
		printk("dt2801: invalid value for 'polarity': %s\n",
		       adc_polarity[bid]);
		return -(EINVAL);
	}

        if (!strcmp(adc_inputmode[bid], "differential")) {
		adc_cf->inputmode  = DT2801_ADC_DIFFERENTIAL;
	} else if (!strcmp(adc_inputmode[bid], "single-ended")) {
		adc_cf->inputmode  = DT2801_ADC_SINGLE_ENDED;
	} else {
		printk("dt2801: invalid value for 'inputmode': %s\n",
		       adc_inputmode[bid]);
		return -(EINVAL);
	}

        if (adc_cf->inputmode == DT2801_ADC_DIFFERENTIAL) {
		adc_cf->nchannels  = 8;
	} else {
		adc_cf->nchannels  = 16;
	}

        return 0;
}

static void dt2801_adc_release(struct inode *inode, struct file *filp)
{
	ADCP* adcp = (ADCP*) filp->private_data;

        if (adcp->dmabuf != NULL) {
                kfree(adcp->dmabuf);
        }

        /* adc-usecount ? */
	filp->private_data = adcp->board;
        kfree(adcp);

        dt2801_release(inode, filp);
}

static struct file_operations adc_fops_imm = {
        NULL,                /* seek */
        adc_read_imm,        /* read */
        NULL,                /* write */
        NULL,                /* readdir */
        NULL,                /* select */
        dt2801_adc_ioctl_a,  /* ioctl */
        NULL,                /* mmap */
        dt2801_adc_open,     /* open */
        dt2801_adc_release   /* release */
};

static struct file_operations adc_fops_block = {
        NULL,                /* seek */
        adc_read_block,      /* read */
        NULL,                /* write */
        NULL,                /* readdir */
        NULL,                /* select */
        dt2801_adc_ioctl,    /* ioctl */
        dt2801_adc_mmap,     /* mmap */
        dt2801_adc_open,     /* open */
        dt2801_adc_release   /* release */
};

int dt2801_adc_open(struct inode *inode, struct file *filp)
{
	int minor = MINOR(inode->i_rdev) & 0xc;
	int channel;
	ADCP* adcp;
        Dt2801_t* boardPtr = filp->private_data;

	DT_PRINTKF("minor = %d\n", minor);

	if ((minor < 0 || minor > 15) && minor != DT2801_ADCX) {
		return -(ENODEV);
	}
        if ( (minor & 0x1f) == DT2801_ADCX) {
		channel = -1;
	} else {
		channel = minor & 0xf;
                if (channel < 0 || channel >= boardPtr->adc_state.nchannels) {
			return -(ENODEV);
		}
	}

	DT_PRINTKF("channel = %d\n", channel);

        if (channel == -1 ) {
		/* non-immediate (block-commands) */
		filp->f_op = &adc_fops_block;
	} else {
		/* immediate */
		filp->f_op = &adc_fops_imm;
	}

	adcp = (ADCP*) kmalloc(sizeof(ADCP), GFP_KERNEL);
	if (adcp == NULL) {
		return -(ENOMEM);
	}

        adcp->board = boardPtr;

	adcp->exttrigger = 0;
	adcp->extclock = 0;
	adcp->continous = 0;

        adcp->startch = 0;
        adcp->endch = 0;
        adcp->gain = 0;
        adcp->conv = 0;

	adcp->last = 0;
	adcp->currnt = 0;
	adcp->running = 0;
	adcp->cnt = 0;

	adcp->dmabufsize = 0;
	adcp->dmabuf = NULL;

	filp->private_data = adcp;

        return 0;
}
