/*
 * Device Driver for DataTranslation DT2801
 *
 * Digital-I/O subsystem
 *
 */

#include <linux/errno.h>
#include <linux/malloc.h>
#include <linux/fcntl.h>

#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <asm/dma.h>

#ifndef MODULE
#error "dio.c must be compiled as a module"
#endif

#include "dt2801.h"
#include "dio.h"


static int set_port_mode(Dt2801_t* boardPtr, int port, int mode) 
{
        DIO_state* dio_cf = &boardPtr->dio_state;

	if (port < 0 || port > dio_cf->nports) {
		printk("dt2801: set_port_mode: invalid port #%d\n", port);
		return -(EINVAL);
	}

        switch (mode) {
        case DT2801_DIO_PORTIN:
		dt2801_writecommand(boardPtr, DT_C_SET_DIGIN);
		dt2801_writedata(boardPtr, port);
                dio_cf->ports[port].mode = DT2801_DIO_PORTIN;
		break;
        case DT2801_DIO_PORTOUT:
		dt2801_writecommand(boardPtr, DT_C_SET_DIGOUT);
		dt2801_writedata(boardPtr, port);
                dio_cf->ports[port].mode = DT2801_DIO_PORTOUT;
		break;
        default:
		printk("dt2801: set_port_mode: invalid mode %d\n", mode);
		return -(EINVAL);
	}

	return 0;
}
 
static int do_write(Dt2801_t* boardPtr, int portnr) 
{
        DIO_state* dio_cf = &boardPtr->dio_state;
	DIO_port* port;
	int i, databyte;

	switch (portnr) {
        case DT2801_DIO0:
        case DT2801_DIO1: {
	}
        case DT2801_DIO2: {
	}
	}
        for (i=0; i<dio_cf->nports; i++) {
		port = &dio_cf->ports[i];
		if (port->mode == DT2801_DIO_PORTOUT) {
			databyte = port->lines;
			dt2801_writecommand(boardPtr, DT_C_WRITE_DIG);
			dt2801_writedata(boardPtr, i);
			dt2801_writedata(boardPtr, databyte);
		} else {
			printk("dt2801: do_write: port %d not writable\n", i);
			return -(EINVAL);
		}
	}
	return 0;
}

static int dio_read(struct inode *inode, struct file *filp,
			   char* buffer, int count)
{
	DIOP* diop = (DIOP*) filp->private_data;
	int minor = MINOR(inode->i_rdev);
	int port = diop->port;

	printk("dt2801: dio_read: minor = %d  port = %d.\n", minor, port);


	return 0;
}

static int dio_read2(struct inode *inode, struct file *filp,
		     char* buffer, int count)
{
	DIOP* diop = (DIOP*) filp->private_data;
	int minor = MINOR(inode->i_rdev);
	int port = diop->port;

	printk("dt2801: dio_read2: minor = %d  port = %d.\n", minor, port);


	return 0;
}

static int dio_write(struct inode *inode, struct file *filp,
		     const char* buffer, int count)
{
	DIOP* diop = (DIOP*) filp->private_data;
	Dt2801_t* boardPtr = diop->board;
        int databyte = 0;


	int minor = MINOR(inode->i_rdev);
	int port = diop->port;

	printk("dt2801: dio_write: minor = %d  port = %d.\n", minor, port);

	dt2801_writecommand(boardPtr, DT_C_WRITE_DIG);
	dt2801_writedata(boardPtr, port);
	dt2801_writedata(boardPtr, databyte);

	return count;
}

static int dio_write2(struct inode *inode, struct file *filp,
		      const char* buffer, int count)
{
	DIOP* diop = (DIOP*) filp->private_data;
	Dt2801_t* boardPtr = diop->board;
        int databyte = 0;


	int minor = MINOR(inode->i_rdev);
	int port = diop->port;

	printk("dt2801: dio_write2: minor = %d  port = %d.\n", minor, port);

	dt2801_writecommand(boardPtr, DT_C_WRITE_DIG);
	dt2801_writedata(boardPtr, port);
	dt2801_writedata2(boardPtr, databyte);

	return count;
}

static int dio_ioctl(struct inode *inode, struct file *filp,
			    unsigned int cmd, unsigned long arg)
{
	DIOP* diop = (DIOP*) filp->private_data;
	Dt2801_t* boardPtr = diop->board;

	int minor = MINOR(inode->i_rdev);
	int port = diop->port;

	printk("dt2801: dio_ioctl: minor = %d  port = %d.\n", minor, port);

	switch (cmd) {
	default:
		return dt2801_do_ioctl(inode, filp, cmd, arg, boardPtr);
	}
	return 0;
}

static int dio_ioctl_g(struct inode *inode, struct file *filp,
			    unsigned int cmd, unsigned long arg)
{
	DIOP* diop = (DIOP*) filp->private_data;
	Dt2801_t* boardPtr = diop->board;

	int minor = MINOR(inode->i_rdev);
	int port = diop->port;

	printk("dt2801: dio_ioctl_g: minor = %d  port = %d.\n", minor, port);

	switch (cmd) {
	case DT2801_IOCTL_DIO_SET_LINES: {
	}
	break;
	default:
		return dt2801_do_ioctl(inode, filp, cmd, arg, boardPtr);
	}
	return 0;
}



/* After board-reset or power-up,
   set both ports to INPUT */
int dt2801_dio_reset(Dt2801_t* boardPtr)
{
        DIO_state* dio_cf = &boardPtr->dio_state;
        int i;

        for (i=0; i<dio_cf->nports; i++) {
		dio_cf->ports[i].mode = DT2801_DIO_PORTIN;
	}

	return 0;
}

/* Called by 'cleanup_module'
   May be called even if 'init' has not  !!!!!! */
int dt2801_dio_cleanup(Dt2801_t* boardPtr)
{

	return 0;
}

void dt2801_dio_info(Dt2801_t* boardPtr)
{
        DIO_state* dio_cf = &boardPtr->dio_state;
        int bid = boardPtr->boardid;

	printk("dt2801:(%d) DIO: %d ports\n", bid, dio_cf->nports);
}

int dt2801_dio_init(Dt2801_t* boardPtr)
{
        DIO_state* dio_cf = &boardPtr->dio_state;
        int i;

        dio_cf->usecount = 0;
        dio_cf->nports = DT2801_DIO_NPORTS;

        for (i=0; i<dio_cf->nports; i++) {
		dio_cf->ports[i].usecount = 0;
		dio_cf->ports[i].mode = DT2801_DIO_PORTUNDEF;
		dio_cf->ports[i].lines = 0;
	}

	return 0;
}

static void dio_release(struct inode *inode, struct file *filp)
{
	DIOP* diop = (DIOP*) filp->private_data;
	Dt2801_t* boardPtr = diop->board;
        DIO_state* dio_cf = &boardPtr->dio_state;

	int i;

	printk("dt2801: dio_release\n");
	dio_cf->ports[diop->port].usecount--;

	filp->private_data = diop->board;
        kfree(diop);
	for (i=0; i<dio_cf->nports; i++) {
		printk("dt2801: dio_usecount[%d] = %d.\n", 
		       i, dio_cf->ports[i].usecount);
	}
	dt2801_release(inode, filp);
}

static void dio_release2(struct inode *inode, struct file *filp)
{
	DIOP* diop = (DIOP*) filp->private_data;
	Dt2801_t* boardPtr = diop->board;
        DIO_state* dio_cf = &boardPtr->dio_state;

	int i;

	printk("dt2801: dio_release2\n");
	dio_cf->ports[0].usecount--;
	dio_cf->ports[1].usecount--;

	filp->private_data = diop->board;
        kfree(diop);
	for (i=0; i<dio_cf->nports; i++) {
		printk("dt2801: dio_usecount[%d] = %d.\n", 
		       i, dio_cf->ports[i].usecount);
	}
	dt2801_release(inode, filp);
}

static struct file_operations dio_fops_r = {
	NULL,                /* seek */
	dio_read,            /* read */
	NULL,                /* write */
	NULL,                /* readdir */
	NULL,                /* select */
	NULL,                /* ioctl */
	NULL,                /* mmap */
	dt2801_dio_open,     /* open */ 
	dio_release          /* release */
};

static struct file_operations dio_fops_w = {
	NULL,                /* seek */
	NULL,                /* read */
	dio_write,           /* write */
	NULL,                /* readdir */
	NULL,                /* select */
	NULL,                /* ioctl */
	NULL,                /* mmap */
	dt2801_dio_open,     /* open */ 
	dio_release          /* release */
};

static struct file_operations dio_fops_r2 = {
	NULL,                /* seek */
	dio_read2,           /* read */
	NULL,                /* write */
	NULL,                /* readdir */
	NULL,                /* select */
	NULL,                /* ioctl */
	NULL,                /* mmap */
	dt2801_dio_open,     /* open */ 
	dio_release2         /* release */
};

static struct file_operations dio_fops_w2 = {
	NULL,                /* seek */
	NULL,                /* read */
	dio_write2,          /* write */
	NULL,                /* readdir */
	NULL,                /* select */
	NULL,                /* ioctl */
	NULL,                /* mmap */
	dt2801_dio_open,     /* open */ 
	dio_release2         /* release */
};

/* A D-I/O port can be opened either for reading or writing
   but not read-write. */
int dt2801_dio_open(struct inode *inode, struct file *filp)
{
        int stat = 0;
	int minor = MINOR(inode->i_rdev) & 0xc;
	int mode;
        int i;

	Dt2801_t* boardPtr = filp->private_data;
        DIO_state* dio_cf = &boardPtr->dio_state;

	DIOP* diop;

	printk("dt2801: dio_open: minor = % d.\n", minor);

	diop = (DIOP*) kmalloc(sizeof(DIOP), GFP_KERNEL);
	if (diop == NULL) {
		return -(ENOMEM);
	}

        diop->board = boardPtr;

	mode = filp->f_flags & O_ACCMODE;

	if (mode == O_RDWR) {
		stat = -(EPERM);
		goto fail1;
	}

	switch (minor) {
        case DT2801_DIO0:
        case DT2801_DIO1: {
		int port = minor & 0x3;

		printk("dt2801: dio_open: port = % d.\n", port);

		if (mode == O_RDONLY) {
			if (dio_cf->ports[port].mode != DT2801_DIO_PORTIN) {
				if (dio_cf->ports[port].usecount != 0) {
					stat = -(EBUSY);
					goto fail1;
				}
				stat = set_port_mode(boardPtr, port,
						     DT2801_DIO_PORTIN);
				if (stat <0) { goto fail1; }
			}
			diop->mode = DT2801_DIO_PORTIN;
			filp->f_op = &dio_fops_r;
		} else { /*O_WRONLY */
			if (dio_cf->ports[port].usecount != 0) {
				stat = -(EBUSY);
				goto fail1;
			}
			stat = set_port_mode(boardPtr, port, DT2801_DIO_PORTOUT);
			if (stat <0) { goto fail1; }
			diop->mode = DT2801_DIO_PORTOUT;
			filp->f_op = &dio_fops_w;
		} 
		dio_cf->ports[port].usecount++;
		
		diop->port = port;
		break;
        }
        case DT2801_DIO2: {
		int port = minor & 0x3; 

		if (mode == O_RDONLY) {
			for (i=0; i<dio_cf->nports; i++) {
				if (dio_cf->ports[i].mode != DT2801_DIO_PORTIN) {
				    if (dio_cf->ports[i].usecount != 0) {
					stat = -(EBUSY);
					goto fail1;
				    }
				    stat = set_port_mode(boardPtr, i,
							 DT2801_DIO_PORTIN);
				    if (stat <0) { goto fail1; }
				}
			}
			diop->mode = DT2801_DIO_PORTIN;
			filp->f_op = &dio_fops_r2;
		} else { /*O_WRONLY */
			for (i=0; i<dio_cf->nports; i++) {
				if (dio_cf->ports[i].usecount != 0) {
					stat = -(EBUSY);
					goto fail1;
				}
			}
			for (i=0; i<dio_cf->nports; i++) {
				stat = set_port_mode(boardPtr, i,
						     DT2801_DIO_PORTOUT);
				if (stat <0) { goto fail1; }
			}
			diop->mode = DT2801_DIO_PORTOUT;
			filp->f_op = &dio_fops_w2;
		} 
		dio_cf->ports[0].usecount++;
		dio_cf->ports[1].usecount++;
		
		diop->port = port;

		break;
	}
        case DT2801_DIOX: {
		printk("dt2801: dio_open: generic device not implemented.\n");
		stat = -(ENODEV);
		goto fail1;

		/* filp->f_op->ioctl = dio_ioctl_g; */

		break;
	}
        default: {
		int line = minor & 0xf;
		int port;

		printk("dt2801: dio_open: line = % d.\n", line);
		if (line <0 || line > dio_cf->nports*DT2801_DIO_PORTSIZE) {
			stat = -(ENODEV);
			goto fail1;
		}
		port = line/DT2801_DIO_PORTSIZE;
		printk("dt2801: dio_open: port = % d.\n", port);
		if (mode == O_RDONLY) {
			if (dio_cf->ports[port].mode != DT2801_DIO_PORTIN) {
				if (dio_cf->ports[port].usecount != 0) {
					stat = -(EBUSY);
					goto fail1;
				}
				stat = set_port_mode(boardPtr, port,
						     DT2801_DIO_PORTIN);
				if (stat <0) { goto fail1; }
			}
			diop->mode = DT2801_DIO_PORTIN;
			filp->f_op = &dio_fops_r2; /* XXX */
		} else { /*O_WRONLY */
			if (dio_cf->ports[port].usecount != 0) {
				stat = -(EBUSY);
				goto fail1;
			}
			stat = set_port_mode(boardPtr, port, DT2801_DIO_PORTOUT);
			if (stat <0) { goto fail1; }
			diop->mode = DT2801_DIO_PORTOUT;
			filp->f_op = &dio_fops_w2; /* XXX */
		}

		dio_cf->ports[port].usecount++;
		
		diop->port = port;
	}
	}

	filp->private_data = diop;

	for (i=0; i<dio_cf->nports; i++) {
		printk("dt2801: dio_usecount[%d] = %d\n.", 
		       i, dio_cf->ports[i].usecount);
	}
	return 0;

fail1:
	kfree(diop);
        return stat;
}


