/*
 * Device Driver for DataTranslation DT2801
 *
 * Linux kernel 2.0.x
 *
 */

static char rcs_revision[] = "$Revision: 0.6.0 $";


#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/tty.h>
#include <linux/signal.h>
#include <linux/malloc.h>
#include <linux/timer.h>
#include <linux/ioport.h>

#include <asm/dma.h>
#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>

#ifdef MODULE
#include <linux/module.h>
#else
#error "dt2801 must be compiled as a module"
#endif

#include "dt2801.h"
#include "adc.h"
#include "dac.h"
#include "dio.h"


static Dt2801_t* boardlist;
static nboards = 0;
/* An element corresponds to a board 
   It will be initialized with pointers into 'boardlist' */
static Dt2801_t* boards[DT2801_BOARDS_MAX] = {NULL, NULL, NULL, NULL};


/* Typeid's for the different boards of the DT2801-series
   (taken from the test-software, that comes with the board)
   DT2801	  0x09
   DT2801-A     0x52
   DT2801/5716A 0x82
   DT2805	  0x12
   DT2805/5716A 0x92
   DT2808	  0x20
   DT2818	  0xA2
   DT2809	  0xB0
*/


static int dt_typeids[DT2801_MAX_TYPES] = {
	DT2801_T_2801,
	DT2801_T_2801A,
	DT2801_T_2801_5716A,
	DT2801_T_2805,
	DT2801_T_2805_5716A,
	DT2801_T_2808,
	DT2801_T_2818,
	DT2801_T_2809,
};

static char* dt_typenames[DT2801_MAX_TYPES+1] = {
   "DT2801",
   "DT2801-A",
   "DT2801/5716A",
   "DT2805",
   "DT2805/5716A",
   "DT2808",
   "DT2818",
   "DT2809",
   "unknown type"
};

static char* typeid2name(int typeid)
{
	int i;
	for (i=0; i<DT2801_MAX_TYPES; i++) {
		if (dt_typeids[i] == typeid) {
			return dt_typenames[i];
		}
	}
	return dt_typenames[i];
}

static void dt2801_timer_interrupt(unsigned long bd)
{
	Dt2801_t* boardPtr = boards[bd];
	if (!boardPtr->dt_timer_int)  return;

	wake_up_interruptible(&boardPtr->waitq);
}

static void dt2801_start_timer(Dt2801_t* boardPtr)
{
	init_timer(&boardPtr->dt_timer_list);
	boardPtr->dt_timer_list.expires = jiffies + DT2801_TIMERCYCLES;
	boardPtr->dt_timer_list.function = dt2801_timer_interrupt;
	boardPtr->dt_timer_list.data = boardPtr->boardid;
	boardPtr->dt_timer_int = 1;
	add_timer(&boardPtr->dt_timer_list);
}

void dt2801_wait(Dt2801_t* boardPtr)
{
	dt2801_start_timer(boardPtr);
 
	interruptible_sleep_on(&boardPtr->waitq);
	boardPtr->dt_timer_int = 0;
	del_timer(&boardPtr->dt_timer_list);
}

/* Stop board, and do a dummy read to clear DataOutReady-bit */
int dt2801_do_stop(Dt2801_t* boardPtr)
{
	int stat, ignored;

	if (boardPtr->fake == 1) {
		return 0;
	}

	stat = inb_p(dt_status(boardPtr));
	/* printk("dt2801: before stop: status = 0x%0x\n", stat); */
	outb_p(DT_C_STOP, dt_command(boardPtr));
	/* sleep a bit ? Wait for ready flag. */

		dt2801_wait(boardPtr);
	ignored = inb_p(dt_data_in(boardPtr));

	stat = inb_p(dt_status(boardPtr));
	/* printk("dt2801: after stop: status = 0x%0x\n", stat); */
		dt2801_wait(boardPtr);
	stat = inb_p(dt_status(boardPtr));
	/* printk("dt2801: after stop: status = 0x%0x\n", stat); */

	return 0;
}

int dt2801_do_reset(Dt2801_t* boardPtr)
{
	int stat, boardtype;

	if (boardPtr->fake == 1) {
		return 0;
	}

	stat = inb_p(dt_status(boardPtr));
	/* printk("dt2801: before reset: status = 0x%0x\n", stat); */
	/* stat != 0 if DataInFull-Flag is cleared */
	stat= ((stat)^DT_S_DATA_IN_FULL) & DT_S_DATA_IN_FULL;
        if (!stat) {
		dt2801_wait(boardPtr);
		stat = inb_p(dt_status(boardPtr));
		stat= ((stat)^DT_S_DATA_IN_FULL) & DT_S_DATA_IN_FULL;
		if (!stat) {
			printk("dt2801: do_reset timed out. "
			       "DataInFull not cleared\n");
			printk("dt2801: status = 0x%0x\n", stat);
		        return -1;
		}
	}

	stat = inb_p(dt_status(boardPtr));
	/* printk("dt2801: rst-1 status = 0x%0x\n", stat); */
	/* stat != 0 if Ready-Flag is set */
	stat = stat & DT_S_READY;
        if (!stat) {
		dt2801_wait(boardPtr);
		stat = inb_p(dt_status(boardPtr));
		stat = stat & DT_S_READY;
		if (!stat) {
			printk("dt2801: do_reset timed out. "
			       "Ready not set\n");
			printk("dt2801: status = 0x%0x\n", stat);
		        return -1;
		}
	}
	outb_p(DT_C_RESET, dt_command(boardPtr));

	/* This is basically readdata(), but completes if
	   Ready OR DataOutReady are set. */
	stat = inb_p(dt_status(boardPtr));
	/* printk("dt2801: rst-2 status = 0x%0x\n", stat); */
	stat = stat & (DT_S_DATA_OUT_READY);
        if (!stat) {
		dt2801_wait(boardPtr);
		stat = inb_p(dt_status(boardPtr));
	        /* printk("dt2801: rst-3 status = 0x%0x\n", stat); */
		stat = stat & ( DT_S_DATA_OUT_READY);
		if (!stat) {
			printk("dt2801: do_reset timed out. "
			       "Ready not set\n");
			printk("dt2801: status = 0x%0x\n", stat);
		        return -1;
		}
	}
		dt2801_wait(boardPtr);
	boardtype = inb_p(dt_data_in(boardPtr));
		dt2801_wait(boardPtr);
	stat = inb_p(dt_status(boardPtr));
	/* printk("dt2801: after reset: status = 0x%0x\n", stat); */

	return boardtype;
}


/* These are the low-level routines:
   writecommand: write a command to the board
   writedata: write data byte
   readdata: read data byte

   The check the status. If the board is not ready the sleep by
   calling dt2801_wait() and try again. If the board is still not
   ready they return with a value != 0 */


/* Only checks DataOutReady-flag, not the Ready-flag as it is done
   in the examples of the manual. I don't see why this should be
   necessary. */
int dt2801_readdata(Dt2801_t* boardPtr, int* data)
{
	static int fake_data = 0;
	int stat = 0;

	if (boardPtr->fake != 0) {
		fake_data = (fake_data + 100)%4096;
		*data = fake_data;
		return 0;
	}

	stat = inb_p(dt_status(boardPtr));
	if (stat & DT_S_COMPOSITE_ERROR) {
            return -1;
	}
	stat = stat & DT_S_DATA_OUT_READY;

	if (!stat) {
		dt2801_wait(boardPtr);

		stat = inb_p(dt_status(boardPtr));
		stat = stat & DT_S_DATA_OUT_READY;

		if (!stat) {
			printk("dt2801: readdata timed out. "
			       "DataOutReady not set\n");
			return -1;
		}

	}
	*data = inb_p(dt_data_in(boardPtr));

	return 0;
}

int dt2801_readdata2(Dt2801_t* boardPtr, int *data)
{
	static int fake_data = 0;
	int lb, hb;

	if (boardPtr->fake != 0) {
		fake_data = (fake_data + 1)%10;
		*data = fake_data;
		return 0;
	}

	if (dt2801_readdata(boardPtr, &lb) != 0) {
		return -1;
	}
	if (dt2801_readdata(boardPtr, &hb) != 0) {
		return -2;
	}

	*data =  (hb<<8)+lb;
	return 0;
}

int dt2801_writedata(Dt2801_t* boardPtr, unsigned int data)
{
	int stat = 0;

	if (boardPtr->fake != 0) {
		return 0;
	}

	stat = inb_p(dt_status(boardPtr));
	if (stat & DT_S_COMPOSITE_ERROR) {
            return -1;
	}
	stat = ((stat)^DT_S_DATA_IN_FULL) & DT_S_DATA_IN_FULL;
	if (!stat) {
		dt2801_wait(boardPtr);
		stat = inb_p(dt_status(boardPtr));
		stat = ((stat)^DT_S_DATA_IN_FULL) & DT_S_DATA_IN_FULL;

		if (!stat) {
			printk("dt2801: writedata timed out. "
				"DataInFull not cleared\n");
			return -1;
		}
	}
	outb_p(data & 0xff, dt_data_out(boardPtr));

	return 0;
}

int dt2801_writedata2(Dt2801_t* boardPtr, unsigned int data)
{

	if ( dt2801_writedata(boardPtr,  data & 0xff) != 0) {
		return -1;
	}
	if ( dt2801_writedata(boardPtr,  (data >> 8) & 0xff ) != 0) {
		return -1;
	}

	return 0;
}

int dt2801_writecommand(Dt2801_t* boardPtr, int command)
{
	int stat;

	if (boardPtr->fake != 0) {
		return 0;
	}

	stat = inb_p(dt_status(boardPtr));
	if (stat & DT_S_COMPOSITE_ERROR) {
        	printk("dt2801: writecommand : warning composite-error\n");
	}
	stat = ((stat)^DT_S_DATA_IN_FULL) & DT_S_DATA_IN_FULL;

	/* stat != 0 if DataInFull-Flag is cleared */
	if (!stat) {
		dt2801_wait(boardPtr);

		stat = inb_p(dt_status(boardPtr));
		stat = ((stat)^DT_S_DATA_IN_FULL) & DT_S_DATA_IN_FULL;

		if (!stat) {
			printk("dt2801: writecommand timed out. "
			       "DataInFull not cleared\n");
			return -1;
		}
	}

	stat = inb_p(dt_status(boardPtr));
	stat = stat & DT_S_READY;

	if (!stat) {
		dt2801_wait(boardPtr);

		stat = inb_p(dt_status(boardPtr));
		stat = stat & DT_S_READY;

		if (!stat) {
			printk("dt2801: writecommand timed out."
			       "Ready not set\n");
			return -1;
		}

	}
	outb_p(command, dt_command(boardPtr));
	return 0;
}

/* Check composite-error-flag. If set read error-register and clear
 */

/* Maybe reset board etc, if we fail to determine the error-code */
int dt2801_checkforerror(Dt2801_t* boardPtr)
{
	int stat;
	int err;

	if (boardPtr->fake == 1) {
		return 0;
	}

	stat = inb_p(dt_status(boardPtr));

        if (stat & DT_S_COMPOSITE_ERROR) {

		/* printk("dt2801: checkforerror: status = 0x%0x.\n", stat); */
		stat = dt2801_writecommand(boardPtr, DT_C_READ_ERRREG);
                if (stat != 0) {
			printk("dt2801: checkforerror: failed to read"
			       " error-reg\n");
			printk("dt2801: reload the driver.\n");
		}
		stat = dt2801_readdata2(boardPtr, &err);
                if (stat != 0) {
			printk("dt2801: checkforerror: failed to read"
			       " error-data\n");
			printk("dt2801: reload the driver.\n");
		}
		printk("dt2801: Error #%d.\n", err);

		/* Clear error-register and composite-error flag */
		stat = dt2801_writecommand(boardPtr, DT_C_CLEAR_ERR);
		if (stat != 0) {
			printk("dt2801: checkforerror: clear-error failed\n");
		}
			
		return err;
	}

        return 0;
}

int dt2801_wait_cmd_completed(Dt2801_t* boardPtr) {
	int stat;

	if (boardPtr->fake == 1) {
		return 0;
	}

	stat = inb_p(dt_status(boardPtr));
	/* printk("dt2801: cmd-1: status = 0x%0x\n", stat); */
	if ( (stat & DT_S_READY) && !(stat & DT_S_COMPOSITE_ERROR)) {
		return 0;
	}
	dt2801_wait(boardPtr);
	stat = inb_p(dt_status(boardPtr));
	/* printk("dt2801: cmd-2: status = 0x%0x\n", stat); */
	if ( (stat & DT_S_READY) && !(stat & DT_S_COMPOSITE_ERROR)) {
		return 0;
	}
	printk("dt2801: cmd not successfully completed\n");

	return -1;
}

int dt2801_clearerror(Dt2801_t* boardPtr) {
	int stat;

	if (boardPtr->fake == 1) {
		return 0;
	}

	stat = inb_p(dt_status(boardPtr));
	/* printk("dt2801: clerrerr-1: status = 0x%0x\n", stat); */
	stat = dt2801_writecommand(boardPtr, DT_C_CLEAR_ERR);
        if (stat != 0) {
		dt2801_checkforerror(boardPtr);
		return -1;
	}
	stat = dt2801_wait_cmd_completed(boardPtr);
        if (stat != 0) {
		dt2801_checkforerror(boardPtr);
		return -1;
	}

	return 0;
}

int dt2801_reset(Dt2801_t* boardPtr)
{
	int stat;

	stat = dt2801_do_reset(boardPtr);
        if (stat < 0) {
               printk("dt2801: reset: do_reset failed\n");
        }

	/* defaults after RESET or power-on */
	dt2801_adc_reset(boardPtr);
	dt2801_dac_reset(boardPtr);
	dt2801_dio_reset(boardPtr);

	boardPtr->clock.ticks = 200;  /* 200 ticks =  1ms for DT2801 */

	return 0;
}

static int dt2801_setclock(Dt2801_t* boardPtr, int period)
{
	int stat, err;

	if (boardPtr->fake == 1) {
		return 0;
	}

	stat = inb_p(dt_status(boardPtr));
        /* printk("dt2801: setclock status = 0x%0x\n", stat); */
	if (stat & DT_S_COMPOSITE_ERROR) {
		err =  dt2801_checkforerror(boardPtr);
                printk("dt2801: setclock err = 0x%0x\n", err);
                return -1;
        }
        /* printk("dt2801: setclock %d\n", period); */
	stat = dt2801_writecommand(boardPtr, DT_C_SET_CLOCK);
        if (stat != 0) {
		err =  dt2801_checkforerror(boardPtr);
		return -1;
	}
	dt2801_writedata2(boardPtr, period);
        if (stat != 0) {
		err =  dt2801_checkforerror(boardPtr);
		return -1;
	}

	stat = dt2801_wait_cmd_completed(boardPtr);
        if (stat != 0) {
		err =  dt2801_checkforerror(boardPtr);
		return -1;
	}

	return stat;
}

int dt2801_ioctl(struct inode *inode, struct file *filp,
		    unsigned int cmd, unsigned long arg)
{
	return 0;
}

int dt2801_do_ioctl(struct inode *inode, struct file *filp,
		    unsigned int cmd, unsigned long arg, Dt2801_t* boardPtr)
{

	switch (cmd) {
	case DT2801_IOCTL_STOP: {
		dt2801_do_stop(boardPtr);
	}
	break;

	case DT2801_IOCTL_SETCLOCK: {
		int cl;
		cl = (int) arg;
		
		if ( dt2801_setclock(boardPtr, cl) != 0) {
			return -1;
		}
		boardPtr->clock.ticks = cl;
	}
	break;
		
	default:
		printk("dt2801: unknown ioctl %u.\n", cmd);
		return -EINVAL;
	}
	
	return 0;
}




/* called by the dt281_xxx_release */
void dt2801_release(struct inode *inode, struct file *filp)
{
	Dt2801_t* boardPtr = filp->private_data;
	boardPtr->usecount--;

        MOD_DEC_USE_COUNT;

	DT_PRINTKF("%d\n", boardPtr->usecount);
}

int dt2801_open(struct inode *inode, struct file *filp)
{
	int stat = 0;
	int minor = MINOR(inode->i_rdev);
	int mode = filp->f_flags & O_ACCMODE;
	int board = DT2801_BOARD(minor);
	Dt2801_t* boardPtr;


	MOD_INC_USE_COUNT;

        if (board <0 || board >= nboards) {
		printk("dt2801_open: no board %d\n", board);
		stat = -(ENODEV);
		MOD_DEC_USE_COUNT;
	        return stat;
	}

	boardPtr = boardlist + board;

	/* stat = inb_p(dt_status(boardPtr));
	   printk("dt2801: open: status = 0x%0x\n", stat); */

        if (boardPtr->firstopen == 1) {
         	if (dt2801_do_reset(boardPtr) < 0) {
	        	printk("dt2801_open: do_reset failed!?\n");
			stat = -(EIO);
			goto fail1;
	        }
		boardPtr->firstopen = 0;
        }


	/* pass the pointer to the board-struct to 'open' */
	filp->private_data = boardPtr;

	if (DT2801_IS_ADC(minor) || DT2801_IS_DAC(minor)) {
		if (DT2801_IS_ADC(minor) && mode == O_RDONLY) {
			stat = dt2801_adc_open(inode, filp);
		} else if (DT2801_IS_DAC(minor) && mode == O_WRONLY) {
			stat = dt2801_dac_open(inode, filp);

		} else {
			printk("dt2801_open: minor = %d, bad AD/DA mode.\n",
			       minor);
			stat = -(EPERM);
		}

	} else if (DT2801_IS_DIO(minor)) {
		stat = dt2801_dio_open(inode, filp);
   
	} else {
		printk("dt2801_open: invalid minor = 0x%02x.\n", minor);
		stat = -(ENODEV);
	}

fail1:
	if (stat < 0) {
		MOD_DEC_USE_COUNT;
	        return stat;
	}
	
        boardPtr->usecount++;
	DT_PRINTKF("%d\n", boardPtr->usecount);

	return stat;

}

struct file_operations dt2801_fops = {
	NULL,            /* seek */
	NULL,            /* read */
	NULL,            /* write */
	NULL,            /* readdir */
	NULL,            /* select */
	dt2801_ioctl,    /* ioctl */
	NULL,            /* mmap */
	dt2801_open,     /* open */
	dt2801_release   /* release */
};

#ifdef MODULE    /* only module version */


/*
 These can be set from the commandline, when loading the module.
*/

/* 0: normal operation, 1: simulate the hardware */
static int fake[DT2801_BOARDS_MAX] = {0, 0 ,0, 0};

/* 0 will result in a dynamically asigned number */
static int major = DT2801_MAJOR;

/* 0 means: no board */
static int io[DT2801_BOARDS_MAX] = {DT2801_BASEADDR, 0x0 ,0x0, 0x0};
/* -1 means: board  will not use dma */
static int dma[DT2801_BOARDS_MAX] = {3, -1,-1, -1};

int init_module(void)
{
	Dt2801_t *boardPtr;
        int i, err, stat = 0;

	printk("dt2801: DataTranslation-DT2801 driver, %s\n", rcs_revision);

        nboards = 0;
	for (i=0; i<DT2801_BOARDS_MAX; i++) {
                int baseaddr = io[i];
                int dmachannel = dma[i];

		if (baseaddr == 0) break;
		if (baseaddr < 0x2 || baseaddr > 0x3fe) {
			printk("dt2801: invalid base address: %d\n", baseaddr);
			return -(EINVAL);
		}

		if (dmachannel != -1) {
			if (dmachannel<1 || dmachannel > 3) {
				printk("dt2801: invalid dmachannel: %d\n",
				       dmachannel);
				return -(EINVAL);
			}
                }
                nboards++;
	}

	boardlist = (Dt2801_t *)kmalloc( sizeof(Dt2801_t) * nboards,
					    GFP_KERNEL);
	if (boardlist == NULL) {
		return -(ENOMEM);
	}

	for (i=0, boardPtr=boardlist; i<nboards; i++, boardPtr++) {

		boards[i] = boardPtr;

		boardPtr->typeid = 0;
		boardPtr->typename = NULL;

		boardPtr->boardid = i;
		boardPtr->fake = fake[i];
		
		boardPtr->dmachannel = -1;
		boardPtr->baseaddr = 0;

		boardPtr->inuse = 0;
		boardPtr->usecount = 0;

		boardPtr->inerror = 0;
		boardPtr->firstopen = 1;
		boardPtr->status = -1;
		boardPtr->dma_running = 0;
		boardPtr->exttrigger_expected = 0;

		boardPtr->dt_timer_int = 0;
		boardPtr->waitq = NULL;
	}

	for (i=0, boardPtr=boardlist; i<nboards; i++, boardPtr++) {
		int dmachannel = dma[i];
		int baseaddr = io[i];

		if (dmachannel != -1) {
			if ( request_dma(dmachannel, "dt2801") ) {
				printk("dt2801_init: Cannot request DMA %d"
				       " for board %d.\n",
				       i,dmachannel);
				boardPtr->dmachannel = -1;
			} else {
				boardPtr->dmachannel = dmachannel;
			}
		}


		if (check_region(baseaddr, DT2801_IOSIZE) != 0) {
			printk("dt2801: init_module: " 
			       "I/O ports 0x%x to 0x%x already in use.\n",
			       baseaddr, baseaddr + DT2801_IOSIZE-1);
			stat = -(EINVAL);
			goto fail_major;
		}
		request_region(baseaddr, DT2801_IOSIZE, "dt2801");
		boardPtr->baseaddr = baseaddr;
	}

	err = register_chrdev(major, "dt2801", &dt2801_fops);
	if (err < 0) {
		printk("dt2801: init_module: "
		       "Cannot register to major device %d.\n", major);
		goto fail_major;
	}
	if (major == 0) {
		major = err; /* dynamic major # */
	}


	/* Stop to make sure the board can accept the next command */
	for (i=0, boardPtr=boardlist; i<nboards; i++, boardPtr++) {
		dt2801_do_stop(boardPtr); 

		stat = dt2801_do_reset(boardPtr);
		if (stat < 0) {
			printk("dt2801_init: board (%d) do_reset failed!?\n",
			       i);
			goto fail3;
		} else {
			boardPtr->typeid = stat;
			boardPtr->typename = typeid2name(boardPtr->typeid);
		}
	}

	/* Initialize Clock, A/D, D/A, D-I/O */

	for (i=0, boardPtr=boardlist; i<nboards; i++, boardPtr++) {
		boardPtr->clock.min = 2;
		boardPtr->clock.max = 0xffff;

		if (dt2801_adc_init(boardPtr) != 0) {
			printk("dt2801_adc_init: failed\n");
			stat = -1;
			goto error;
		}
		if (dt2801_dac_init(boardPtr) != 0) {
			printk("dt2801_dac_init: failed\n");
			stat = -1;
			goto error;
		}
		if (dt2801_dio_init(boardPtr) != 0) {
			printk("dt2801_dio_init: failed\n");
			stat = -1;
			goto error;
		}
	}
		
	/* Reset the board, via dt2801_reset(),
	   which will call the dt2801_xxx_reset() routines. */
	for (boardPtr = boardlist; boardPtr < boardlist+nboards; boardPtr++) {
		if (dt2801_reset(boardPtr) != 0) {
			printk("dt2801_init: reset not ok.\n");
			goto error;
		}
	}
	
	printk("dt2801: major(%d).\n", major);
        if (nboards == 0) {
		printk("dt2801: NO boards.\n");
	} else {
		printk("dt2801: #of boards: %d.\n", nboards);
		for (boardPtr = boardlist; boardPtr < boardlist + nboards;
		     boardPtr++) {
			printk("dt2801:(%d) %-15s fake(%d), io(0x%lx), dma(%d)\n",
			       boardPtr->boardid,
			       boardPtr->typename,
			       boardPtr->fake,
			       boardPtr->baseaddr,
			       boardPtr->dmachannel);
			dt2801_adc_info(boardPtr);
			dt2801_dac_info(boardPtr);
			dt2801_dio_info(boardPtr);
		}
	}

	/* A module consisting of more than one object-file will
	   export all symbols (even when declared static). So we
	   explicitly remove the symbols from the table.
	   c.f. /usr/doc/modules/HOWTO-modularize. */
	register_symtab(0);

	/* stat = inb_p(dt_status(boardPtr)); */
	/* printk("dt2801: finally: status = 0x%0x\n", stat); */

	return 0;

error:

	printk("dt2801: error during init, cleaning up.\n");
	for (boardPtr = boardlist; boardPtr < boardlist+nboards; boardPtr++) {
		dt2801_adc_cleanup(boardPtr);
		dt2801_dac_cleanup(boardPtr);
		dt2801_dio_cleanup(boardPtr);
	}

fail3:
	unregister_chrdev(major, "dt2801");

fail_major:
	for (boardPtr = boardlist; boardPtr < boardlist+nboards; boardPtr++) {
		if (boardPtr->baseaddr != 0) {
			release_region(boardPtr->baseaddr, DT2801_IOSIZE);
		}
	}

	for (boardPtr = boardlist; boardPtr < boardlist+nboards; boardPtr++) {
		if (boardPtr->dmachannel != -1) {
			free_dma(boardPtr->dmachannel);
		}
	}

	kfree(boardlist);

        /* If I forgot to set 'stat' in case of an error. Force
	   unloading the module */
        if (stat == 0) { stat = -1; }
	return stat;
}

void cleanup_module(void)
{
        Dt2801_t* boardPtr;

	for (boardPtr = boardlist; boardPtr < boardlist+nboards; boardPtr++) {
		dt2801_adc_cleanup(boardPtr);
		dt2801_dac_cleanup(boardPtr);
		dt2801_dio_cleanup(boardPtr);

		if (boardPtr->dmachannel != -1) {
			free_dma(boardPtr->dmachannel);
		}
		release_region(boardPtr->baseaddr, DT2801_IOSIZE);
	}

	if (unregister_chrdev(major, "dt2801") != 0) {
		printk("dt2801: cleanup_module failed\n");
	} else {
		printk("dt2801: cleanup succeeded\n");
	}

	kfree(boardlist);
}

#endif /* MODULE */
